@@ -1,11 +1,12 @@
 #This file is part of Tryton.  The COPYRIGHT file at the top level of
 #this repository contains the full copyright notices and license terms.
 import os
+from itertools import chain
 import tempfile
 import locale
 from tryton.common import datetime_strftime, \
-        domain_inversion, eval_domain, localize_domain, unlocalize_domain, \
-        merge, inverse_leaf, EvalEnvironment
+        domain_inversion, eval_domain, localize_domain, \
+        merge, inverse_leaf, concat, simplify, EvalEnvironment
 import tryton.common as common
 import time
 import datetime
@@ -23,15 +24,11 @@ class Field(object):
     set: save the value from the server
     set_client: save the value from the widget
     '''
+    _default = None
 
-    def __new__(cls, ctype):
-        klass = TYPES.get(ctype, CharField)
-        return klass
-
-
-class CharField(object):
-
-    _default = ''
+    @staticmethod
+    def get_field(ctype):
+        return TYPES.get(ctype, CharField)
 
     def __init__(self, attrs):
         self.attrs = attrs
@@ -49,7 +46,7 @@ class CharField(object):
         screen_domain = domain_inversion(record.group.domain4inversion,
             self.name, EvalEnvironment(record))
         if isinstance(screen_domain, bool) and not screen_domain:
-            screen_domain = [('id', '=', False)]
+            screen_domain = [('id', '=', None)]
         elif isinstance(screen_domain, bool) and screen_domain:
             screen_domain = []
         attr_domain = record.expr_eval(self.attrs.get('domain', []))
@@ -57,15 +54,10 @@ class CharField(object):
 
     def domain_get(self, record):
         screen_domain, attr_domain = self.domains_get(record)
-        return [localize_domain(screen_domain), attr_domain]
+        return concat(localize_domain(screen_domain), attr_domain)
 
     def validation_domains(self, record):
-        screen_domain, attr_domain = self.domains_get(record)
-        if attr_domain:
-            return (screen_domain, [screen_domain,
-                    unlocalize_domain(attr_domain, self.name)])
-        else:
-            return screen_domain, screen_domain
+        return concat(*self.domains_get(record))
 
     def context_get(self, record):
         context = record.context_get().copy()
@@ -87,27 +79,30 @@ class CharField(object):
             return True
         res = True
         self.get_state_attrs(record)['domain_readonly'] = False
-        inverted_domain, domain = self.validation_domains(record)
+        domain = simplify(self.validation_domains(record))
         if not softvalidation:
             res = res and self.check_required(record)
         if isinstance(domain, bool):
             res = res and domain
-        elif domain == [('id', '=', False)]:
+        elif domain == [('id', '=', None)]:
             res = False
         else:
-            if (isinstance(inverted_domain, list)
-                    and len(inverted_domain) == 1
-                    and inverted_domain[0][1] == '='):
+            if (isinstance(domain, list)
+                    and len(domain) == 1
+                    and domain[0][1] == '='):
                 # If the inverted domain is so constraint that only one value
                 # is possible we should use it. But we must also pay attention
                 # to the fact that the original domain might be a 'OR' domain
                 # and thus not preventing the modification of fields.
-                leftpart, _, value = inverted_domain[0][:3]
+                leftpart, _, value = domain[0][:3]
                 if value is False:
                     # XXX to remove once server domains are fixed
                     value = None
                 setdefault = True
-                original_domain = merge(record.group.domain)
+                if record.group.domain:
+                    original_domain = merge(record.group.domain)
+                else:
+                    original_domain = merge(domain)
                 domain_readonly = original_domain[0] == 'AND'
                 if '.' in leftpart:
                     recordpart, localpart = leftpart.split('.', 1)
@@ -129,7 +124,7 @@ class CharField(object):
         record.value[self.name] = value
 
     def get(self, record):
-        return record.value.get(self.name) or self._default
+        return record.value.get(self.name, self._default)
 
     def get_eval(self, record):
         return self.get(record)
@@ -152,7 +147,7 @@ class CharField(object):
             record.signal('record-changed')
 
     def get_client(self, record):
-        return record.value.get(self.name) or self._default
+        return self.get(record)
 
     def set_default(self, record, value):
         self.set(record, value)
@@ -186,7 +181,14 @@ class CharField(object):
         return {}
 
 
-class SelectionField(CharField):
+class CharField(Field):
+    _default = ''
+
+    def get(self, record):
+        return super(CharField, self).get(record) or self._default
+
+
+class SelectionField(Field):
 
     _default = None
 
@@ -194,18 +196,19 @@ class SelectionField(CharField):
         return record.value.get(self.name)
 
 
-class DateTimeField(CharField):
+class DateTimeField(Field):
 
     _default = None
 
     def set_client(self, record, value, force_change=False):
-        if not isinstance(value, datetime.datetime):
+        if isinstance(value, basestring):
             try:
                 value = datetime.datetime(*time.strptime(value,
                         date_format() + ' ' + self.time_format(record))[:6])
-                value = common.untimezoned_date(value)
             except ValueError:
                 value = self._default
+        if value:
+            value = common.untimezoned_date(value)
         super(DateTimeField, self).set_client(record, value,
             force_change=force_change)
 
@@ -221,17 +224,20 @@ class DateTimeField(CharField):
         return record.expr_eval(self.attrs['format'])
 
 
-class DateField(CharField):
+class DateField(Field):
 
     _default = None
 
     def set_client(self, record, value, force_change=False):
-        if not isinstance(value, datetime.date):
+        if isinstance(value, basestring):
             try:
                 value = datetime.date(*time.strptime(value,
                         date_format())[:3])
             except ValueError:
                 value = self._default
+        elif isinstance(value, datetime.datetime):
+            assert(value.time() == datetime.time())
+            value = value.date()
         super(DateField, self).set_client(record, value,
             force_change=force_change)
 
@@ -242,17 +248,19 @@ class DateField(CharField):
         return ''
 
 
-class TimeField(CharField):
+class TimeField(Field):
 
     _default = None
 
     def set_client(self, record, value, force_change=False):
-        if not isinstance(value, datetime.time):
+        if isinstance(value, basestring):
             try:
                 value = datetime.time(*time.strptime(value,
                         self.time_format(record))[3:6])
             except ValueError:
-                value = None
+                value = self._default
+        elif isinstance(value, datetime.datetime):
+            value = value.time()
         super(TimeField, self).set_client(record, value,
             force_change=force_change)
 
@@ -266,7 +274,7 @@ class TimeField(CharField):
         return record.expr_eval(self.attrs['format'])
 
 
-class FloatField(CharField):
+class FloatField(Field):
     _default = None
     default_digits = (16, 2)
 
@@ -286,7 +294,7 @@ class FloatField(CharField):
                 record.expr_eval(
                     self.attrs.get('digits', self.default_digits)),
                 self.default_digits))
-        shift = int(round(math.log(factor, 10)))
+        shift = int(round(math.log(abs(factor), 10)))
         return (digits[0] + shift, digits[1] - shift)
 
     def convert(self, value):
@@ -346,7 +354,7 @@ class IntegerField(FloatField):
         return super(IntegerField, self).get_client(record, factor=int(factor))
 
 
-class BooleanField(CharField):
+class BooleanField(Field):
 
     _default = False
 
@@ -362,7 +370,7 @@ class BooleanField(CharField):
         return bool(record.value.get(self.name))
 
 
-class M2OField(CharField):
+class M2OField(Field):
     '''
     internal = (id, name)
     '''
@@ -410,7 +418,7 @@ class M2OField(CharField):
         if not rec_name and value >= 0:
             try:
                 result, = RPCExecute('model', self.attrs['relation'], 'read',
-                    [value], ['rec_name'], main_iteration=False)
+                    [value], ['rec_name'])
             except RPCException:
                 return False
             rec_name = result['rec_name'] or ''
@@ -431,12 +439,12 @@ class M2OField(CharField):
 
     def validation_domains(self, record):
         screen_domain, attr_domain = self.domains_get(record)
-        return screen_domain, screen_domain
+        return screen_domain
 
     def domain_get(self, record):
         screen_domain, attr_domain = self.domains_get(record)
-        return [localize_domain(inverse_leaf(screen_domain), self.name),
-            attr_domain]
+        return concat(localize_domain(inverse_leaf(screen_domain), self.name),
+            attr_domain)
 
     def get_state_attrs(self, record):
         result = super(M2OField, self).get_state_attrs(record)
@@ -451,7 +459,7 @@ class O2OField(M2OField):
     pass
 
 
-class O2MField(CharField):
+class O2MField(Field):
     '''
     internal = Group of the related objects
     '''
@@ -470,10 +478,12 @@ class O2MField(CharField):
     def _group_changed(self, group, record):
         if not record.parent:
             return
-        record.parent.modified_fields.setdefault(self.name)
-        self.sig_changed(record.parent)
-        record.parent.validate(softvalidation=True)
-        record.parent.signal('record-changed')
+        # Store parent as it could be removed by validation
+        parent = record.parent
+        parent.modified_fields.setdefault(self.name)
+        self.sig_changed(parent)
+        parent.validate(softvalidation=True)
+        parent.signal('record-changed')
 
     def _group_list_changed(self, group, signal):
         if group.model_name == group.parent.model_name:
@@ -521,9 +531,11 @@ class O2MField(CharField):
             return []
         record_removed = record.value[self.name].record_removed
         record_deleted = record.value[self.name].record_deleted
-        result = [('add', [])]
+        result = []
         parent_name = self.attrs.get('relation_field', '')
+        to_add = []
         to_create = []
+        to_write = []
         for record2 in record.value[self.name]:
             if record2 in record_removed or record2 in record_deleted:
                 continue
@@ -531,18 +543,20 @@ class O2MField(CharField):
                 values = record2.get()
                 values.pop(parent_name, None)
                 if record2.modified and values:
-                    result.append(('write', [record2.id], values))
-                result[0][1].append(record2.id)
+                    to_write.extend(([record2.id], values))
+                to_add.append(record2.id)
             else:
                 values = record2.get()
                 values.pop(parent_name, None)
                 to_create.append(values)
+        if to_add:
+            result.append(('add', to_add))
         if to_create:
             result.append(('create', to_create))
-        if not result[0][1]:
-            del result[0]
+        if to_write:
+            result.append(('write',) + tuple(to_write))
         if record_removed:
-            result.append(('unlink', [x.id for x in record_removed]))
+            result.append(('remove', [x.id for x in record_removed]))
         if record_deleted:
             result.append(('delete', [x.id for x in record_deleted]))
         return result
@@ -590,15 +604,15 @@ class O2MField(CharField):
             if field_names:
                 try:
                     fields = RPCExecute('model', self.attrs['relation'],
-                        'fields_get', list(field_names),
-                        main_iteration=False, context=context)
+                        'fields_get', list(field_names), context=context)
                 except RPCException:
                     return
                 group.load_fields(fields)
 
         if mode == 'list ids':
             for old_record in group:
-                group.remove(old_record, remove=True, signal=False)
+                if old_record.id not in value:
+                    group.remove(old_record, remove=True, signal=False)
             group.load(value)
         else:
             for vals in value:
@@ -634,7 +648,23 @@ class O2MField(CharField):
         self._connect_value(group)
 
     def set_client(self, record, value, force_change=False):
-        pass
+        # domain inversion could try to set id as value
+        if isinstance(value, (int, long)):
+            value = [value]
+
+        previous_ids = self.get_eval(record)
+        self._set_value(record, value)
+        # The order of the ids is not significant
+        if set(previous_ids) != set(value):
+            record.modified_fields.setdefault(self.name)
+            record.signal('record-modified')
+            self.sig_changed(record)
+            record.validate(softvalidation=True)
+            record.signal('record-changed')
+        elif force_change:
+            self.sig_changed(record)
+            record.validate(softvalidation=True)
+            record.signal('record-changed')
 
     def set_default(self, record, value):
         self.set(record, value, _default=True)
@@ -650,21 +680,21 @@ class O2MField(CharField):
         if value and (value.get('add') or value.get('update')):
             context = self.context_get(record)
             fields = record.value[self.name].fields
-            field_names = set(f for v in (
-                    value.get('add', []) + value.get('update', []))
+            values = chain(value.get('update', []),
+                (d for _, d in value.get('add', [])))
+            field_names = set(f for v in values
                 for f in v if f not in fields and f != 'id')
             if field_names:
                 try:
                     fields = RPCExecute('model', self.attrs['relation'],
-                        'fields_get', list(field_names), main_iteration=False,
-                        context=context)
+                        'fields_get', list(field_names), context=context)
                 except RPCException:
                     return False
             else:
                 fields = {}
 
         to_remove = []
-        for record2 in record.value[self.name]:
+        for record2 in record.value.get(self.name, []):
             if not record2.id:
                 to_remove.append(record2)
         if value and value.get('remove'):
@@ -678,9 +708,9 @@ class O2MField(CharField):
 
         if value and (value.get('add') or value.get('update', [])):
             record.value[self.name].add_fields(fields, signal=False)
-            for vals in value.get('add', []):
+            for index, vals in value.get('add', []):
                 new_record = record.value[self.name].new(default=False)
-                record.value[self.name].add(new_record)
+                record.value[self.name].add(new_record, index)
                 new_record.set_on_change(vals)
 
             for vals in value.get('update', []):
@@ -693,7 +723,7 @@ class O2MField(CharField):
 
     def validation_domains(self, record):
         screen_domain, attr_domain = self.domains_get(record)
-        return screen_domain, screen_domain
+        return screen_domain
 
     def validate(self, record, softvalidation=False):
         if self.attrs.get('readonly'):
@@ -723,7 +753,8 @@ class O2MField(CharField):
 
     def domain_get(self, record):
         screen_domain, attr_domain = self.domains_get(record)
-        return [localize_domain(inverse_leaf(screen_domain)), attr_domain]
+        return concat(localize_domain(inverse_leaf(screen_domain), self.name),
+            attr_domain)
 
 
 class M2MField(O2MField):
@@ -732,7 +763,7 @@ class M2MField(O2MField):
         return self.get_eval(record)
 
 
-class ReferenceField(CharField):
+class ReferenceField(Field):
 
     _default = None
 
@@ -780,6 +811,12 @@ class ReferenceField(CharField):
             force_change=force_change)
 
     def set(self, record, value):
+        if record.parent_name == self.name:
+            if record.parent:
+                value = '%s,%s' % (record.group.parent.model_name,
+                    record.parent.id)
+            else:
+                value = None
         if not value:
             record.value[self.name] = self._default
             return
@@ -799,24 +836,24 @@ class ReferenceField(CharField):
             if not rec_name and ref_id >= 0:
                 try:
                     result, = RPCExecute('model', ref_model, 'read', [ref_id],
-                        ['rec_name'], main_iteration=False)
+                        ['rec_name'])
                 except RPCException:
                     return
                 rec_name = result['rec_name']
         elif ref_model:
             rec_name = ''
         else:
-            rec_name = ref_id
+            rec_name = str(ref_id)
         record.value[self.name] = ref_model, ref_id
         record.value[self.name + '.rec_name'] = rec_name
 
 
-class BinaryField(CharField):
+class BinaryField(Field):
 
     _default = None
 
     def get(self, record):
-        result = record.value.get(self.name) or self._default
+        result = record.value.get(self.name, self._default)
         if isinstance(result, basestring):
             try:
                 with open(result, 'rb') as fp:
@@ -854,8 +891,7 @@ class BinaryField(CharField):
             context = record.context_get()
             try:
                 values, = RPCExecute('model', record.model_name, 'read',
-                    [record.id], [self.name], main_iteration=False,
-                    context=context)
+                    [record.id], [self.name], context=context)
             except RPCException:
                 return ''
             _, filename = tempfile.mkstemp(prefix='tryton_')
@@ -865,21 +901,27 @@ class BinaryField(CharField):
         return self.get(record)
 
 
-class DictField(CharField):
+class DictField(Field):
 
     _default = {}
 
+    def get(self, record):
+        return super(DictField, self).get(record) or self._default
+
+    def get_client(self, record):
+        return super(DictField, self).get_client(record) or self._default
+
     def validation_domains(self, record):
         screen_domain, attr_domain = self.domains_get(record)
-        return screen_domain, screen_domain
+        return screen_domain
 
     def domain_get(self, record):
         screen_domain, attr_domain = self.domains_get(record)
-        return [localize_domain(inverse_leaf(screen_domain)), attr_domain]
+        return concat(localize_domain(inverse_leaf(screen_domain)),
+            attr_domain)
 
 TYPES = {
     'char': CharField,
-    'sha': CharField,
     'float_time': FloatField,
     'integer': IntegerField,
     'biginteger': IntegerField,
